@@ -1,6 +1,3 @@
-# vim: tabstop=4 shiftwidth=4 softtabstop=4
-# coding=utf8
-
 # Copyright 2015 Lenovo
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
@@ -15,27 +12,31 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-# This module provides access to SDR offered by a BMC
-# This data is common between 'sensors' and 'inventory' modules since SDR
-# is both used to enumerate sensors for sensor commands and FRU ids for FRU
-# commands
+"""This module provides access to SDR offered by a BMC
 
-# For now, we will not offer persistent SDR caching as we do in xCAT's IPMI
-# code.  Will see if it is adequate to advocate for high object reuse in a
-# persistent process for the moment.
+This data is common between 'sensors' and 'inventory' modules since SDR
+is both used to enumerate sensors for sensor commands and FRU ids for FRU
+commands
 
-# Focus is at least initially on the aspects that make the most sense for a
-# remote client to care about.  For example, smbus information is being
-# skipped for now
+For now, we will not offer persistent SDR caching as we do in xCAT's IPMI
+code.  Will see if it is adequate to advocate for high object reuse in a
+persistent process for the moment.
 
-# This file handles parsing of fru format records as presented by IPMI
-# devices.  This format is documented in the 'Platform Management FRU
-# Information Storage Definition (Document Revision 1.2)
+Focus is at least initially on the aspects that make the most sense for a
+remote client to care about.  For example, smbus information is being
+skipped for now
+
+This file handles parsing of fru format records as presented by IPMI
+devices.  This format is documented in the 'Platform Management FRU
+Information Storage Definition (Document Revision 1.2)
+"""
+
+import struct
+import time
 
 import pyghmi.exceptions as iexc
 import pyghmi.ipmi.private.spd as spd
-import struct
-import time
+
 
 fruepoch = 820454400  # 1/1/1996, 0:00
 
@@ -187,8 +188,8 @@ class FRU(object):
             frusubtype = self.sdr.fru_type_and_modifier & 0xff
             if frutype > 0x10 or frutype < 0x8 or frusubtype not in (0, 1, 2):
                 return
-                #TODO(jjohnson2): strict mode to detect pyghmi and BMC
-                #gaps
+                # TODO(jjohnson2): strict mode to detect pyghmi and BMC
+                # gaps
                 # raise iexc.PyghmiException(
                 #     'Unsupported FRU device: {0:x}h, {1:x}h'.format(frutype,
                 #                                                    frusubtype
@@ -199,7 +200,7 @@ class FRU(object):
                 return
         if self.databytes[0] != 1:
             return
-            #TODO(jjohnson2): strict mode to flag potential BMC errors
+            # TODO(jjohnson2): strict mode to flag potential BMC errors
             # raise iexc.BmcErrorException("Invalid/Unsupported FRU format")
         # Ignore the internal use even if present.
         self._parse_chassis()
@@ -225,22 +226,25 @@ class FRU(object):
             # strictly speaking, \xff should be a y with diaeresis, but
             # erring on the side of that not being very relevant in practice
             # to fru info, particularly the last values
-            retinfo = retinfo.rstrip('\xff\x00 ')
+            # Additionally 0xfe has been observed, which should be a thorn, but
+            # again assuming termination of string is more likely than thorn.
+            retinfo = retinfo.rstrip(b'\xfe\xff\x10\x03\x00 ')
+            retinfo = retinfo.replace(b'\x00', b'')
             if lang in (0, 25):
                 try:
                     retinfo = retinfo.decode('iso-8859-1')
-                except UnicodeDecodeError:
+                except (UnicodeError, LookupError):
                     pass
             else:
                 try:
                     retinfo = retinfo.decode('utf-16le')
-                except UnicodeDecodeError:
+                except (UnicodeDecodeError, LookupError):
                     pass
             # Some things lie about being text.  Do the best we can by
             # removing trailing spaces and nulls like makes sense for text
             # and rely on vendors to workaround deviations in their OEM
             # module
-            retinfo = retinfo.rstrip('\x00 ')
+            # retinfo = retinfo.rstrip(b'\x00 ')
             return retinfo, newoffset
         elif currtype == 1:  # BCD 'plus'
             retdata = ''
@@ -259,7 +263,7 @@ class FRU(object):
         if offset == 0:
             return
         if self.databytes[offset] & 0b1111 != 1:
-            raise iexc.BmcErrorException("Invallid/Unsupported chassis area")
+            raise iexc.BmcErrorException("Invalid/Unsupported chassis area")
         inf = self.info
         # ignore length field, just process the data
         inf['Chassis type'] = enclosure_types[self.databytes[offset + 2]]
