@@ -27,6 +27,8 @@ import socket
 import struct
 import weakref
 
+import six
+
 import pyghmi.constants as pygconst
 import pyghmi.exceptions as pygexc
 import pyghmi.ipmi.oem.lenovo.config as config
@@ -37,7 +39,6 @@ from pyghmi.ipmi import sdr
 import pyghmi.media as media
 import pyghmi.storage as storage
 import pyghmi.util.webclient as webclient
-import six
 
 try:
     from urllib import urlencode
@@ -123,15 +124,16 @@ class IMMClient(object):
         self.ipmicmd = weakref.proxy(ipmicmd)
         self.updating = False
         self.imm = ipmicmd.bmc
-        self.adp_referer = 'https://{0}/designs/imm/index-console.php'.format(
-            self.imm)
+        srv = self.imm
+        if ':' in srv:
+            srv = '[{0}]'.format(srv)
+        self.adp_referer = 'https://imm/designs/imm/index-console.php'
         if ipmicmd.ipmi_session.password:
             self.username = ipmicmd.ipmi_session.userid.decode('utf-8')
             self.password = ipmicmd.ipmi_session.password.decode('utf-8')
         self._wc = None  # The webclient shall be initiated on demand
         self._energymanager = None
         self.datacache = {}
-        self.webkeepalive = None
         self._keepalivesession = None
         self.fwc = None
         self.fwo = None
@@ -188,8 +190,8 @@ class IMMClient(object):
         try:
             self.fwo = self.fwc.get_fw_options(fetchimm=fetchimm)
         except Exception:
-            raise Exception(self.bmcname +
-                            ' failed to retrieve UEFI configuration')
+            raise Exception('%s failed to retrieve UEFI configuration'
+                            % self.bmcname)
         self.fwovintage = util._monotonic_time()
         retcfg = {}
         for opt in self.fwo:
@@ -197,8 +199,8 @@ class IMMClient(object):
                 # Suppress the Avago configuration to be consistent with
                 # other tools.
                 continue
-            if (hideadvanced and self.fwo[opt]['lenovo_protect'] or
-                    self.fwo[opt]['hidden']):
+            if (hideadvanced and self.fwo[opt]['lenovo_protect']
+                    or self.fwo[opt]['hidden']):
                 # Do not enumerate hidden settings
                 continue
             retcfg[opt] = {}
@@ -250,6 +252,15 @@ class IMMClient(object):
                 else:
                     raise pygexc.InvalidParameterValue(
                         '{0} not a known setting'.format(key))
+        self.merge_changeset(changeset)
+        if changeset:
+            try:
+                self.fwc.set_fw_options(self.fwo)
+            finally:
+                self.fwo = None
+                self.fwovintage = 0
+
+    def merge_changeset(self, changeset):
         for key in changeset:
             if isinstance(changeset[key], six.string_types):
                 changeset[key] = {'value': changeset[key]}
@@ -266,8 +277,8 @@ class IMMClient(object):
             newnewvalues = []
             for newvalue in newvalues:
                 newv = re.sub(r'\s+', ' ', newvalue)
-                if (self.fwo[key]['possible'] and
-                        newvalue not in self.fwo[key]['possible']):
+                if (self.fwo[key]['possible']
+                        and newvalue not in self.fwo[key]['possible']):
                     candlist = []
                     for candidate in self.fwo[key]['possible']:
                         candid = re.sub(r'\s+', ' ', candidate)
@@ -290,12 +301,6 @@ class IMMClient(object):
                 self.fwo[key]['new_value'] = newnewvalues[0]
             else:
                 self.fwo[key]['new_value'] = newnewvalues
-        if changeset:
-            try:
-                self.fwc.set_fw_options(self.fwo)
-            except Exception:
-                self.fwo = None
-                raise
 
     def clear_bmc_configuration(self):
         self.ipmicmd.xraw_command(0x2e, 0xcc,
@@ -331,6 +336,8 @@ class IMMClient(object):
             if not isinstance(ret, str):
                 ret = ret.decode('utf-8')
             return ret
+        if propdata[0] == 0x44:  # dword
+            return propdata[1:5]
         else:
             raise Exception('Unknown format for property: ' + repr(propdata))
 
@@ -348,8 +355,9 @@ class IMMClient(object):
                            'password': self.password,
                            'SessionTimeout': 60})
         headers = {'Connection': 'keep-alive',
-                   'Referer': 'https://{0}/designs/imm/index.php'.format(
-                       self.imm),
+                   'Origin': 'https://imm/',
+                   'Host': 'imm',
+                   'Referer': 'https://imm/designs/imm/index.php',
                    'Content-Type': 'application/x-www-form-urlencoded'}
         wc.request('POST', '/data/login', adata, headers)
         rsp = wc.getresponse()
@@ -365,13 +373,19 @@ class IMMClient(object):
                 else:
                     self.uploadtoken = {}
                 wc.set_header('Referer', self.adp_referer)
-                wc.set_header('Host', self.imm)
+                wc.set_header('Host', 'imm')
+                wc.set_header('Origin', 'https://imm/')
                 return wc
 
     @property
     def wc(self):
-        if (not self._wc or (self._wc.vintage and
-                             self._wc.vintage < util._monotonic_time() - 30)):
+        if (not self._wc or (self._wc.vintage
+                             and self._wc.vintage < util._monotonic_time()
+                             - 30)):
+            if not self.updating and self._wc:
+                # in case the existing session is still valid
+                # dispose of the session
+                self.weblogout()
             self._wc = self.get_webclient()
         return self._wc
 
@@ -393,7 +407,12 @@ class IMMClient(object):
     def grab_cacheable_json(self, url, age=30):
         data = self.get_cached_data(url, age)
         if not data:
-            data = self.wc.grab_json_response(url)
+            data, status = self.wc.grab_json_response_with_status(url)
+            if status == 401:
+                self._wc = None
+                data, status = self.wc.grab_json_response_with_status(url)
+            if status != 200:
+                data = {}
             self.datacache[url] = (data, util._monotonic_time())
         return data
 
@@ -405,7 +424,7 @@ class IMMClient(object):
         except KeyError:
             return None
 
-    def upload_media(self, filename, progress=None):
+    def upload_media(self, filename, progress=None, data=None):
         xid = random.randint(0, 1000000000)
         alloc = self.wc.grab_json_response(
             '/data/set',
@@ -419,7 +438,7 @@ class IMMClient(object):
         uploadfields['available'] = alloc['available']
         uploadfields['checksum'] = xid
         ut = webclient.FileUploader(
-            self.wc, '/designs/imm/upload/rp_image_upload.esp', filename,
+            self.wc, '/designs/imm/upload/rp_image_upload.esp', filename, data,
             otherfields=uploadfields)
         ut.start()
         while ut.isAlive():
@@ -470,6 +489,11 @@ class IMMClient(object):
                 for uload in item['images']:
                     if uload['status'] != 0:
                         yield media.Media(uload['filename'])
+            for attached in item.get('urls', []):
+                filename = attached['url']
+                filename = filename.split('/')[-1]
+                url = '/'.join(attached['url'].split('/')[:-1])
+                yield media.Media(filename, url)
 
     def detach_remote_media(self):
         mnt = self.wc.grab_json_response(
@@ -548,22 +572,22 @@ class IMMClient(object):
                         bdata = {}
                         if 'versionStr' in firm and firm['versionStr']:
                             bdata['version'] = firm['versionStr']
-                        if ('releaseDate' in firm and
-                                firm['releaseDate'] and
-                                firm['releaseDate'] != 'N/A'):
+                        if ('releaseDate' in firm
+                                and firm['releaseDate']
+                                and firm['releaseDate'] != 'N/A'):
                             try:
                                 bdata['date'] = self._parse_builddate(
                                     firm['releaseDate'])
                             except ValueError:
                                 pass
-                        yield ('{0} {1}'.format(aname, fname), bdata)
+                        yield '{0} {1}'.format(aname, fname), bdata
                 for fwi in fwu.get('items', []):
                     if fwi.get('key', -1) == adata.get('key', -2):
                         if fwi.get('fw_status', 0) == 2:
                             bdata = {}
                             if 'fw_version_pend' in fwi:
                                 bdata['version'] = fwi['fw_version_pend']
-                            yield('{0} Pending Update'.format(aname), bdata)
+                            yield '{0} Pending Update'.format(aname), bdata
         for disk in self.disk_inventory():
             yield disk
         self.weblogout()
@@ -704,8 +728,8 @@ class IMMClient(object):
                     bdata['pcislot'] = '{0:02x}:{1:02x}'.format(
                         fundata[self.BUSNO], fundata[self.DEVNO])
                     serialdata = fundata.get(self.ADP_SERIALNO, None)
-                    if (serialdata and serialdata != 'N/A' and
-                            '---' not in serialdata):
+                    if (serialdata and serialdata != 'N/A'
+                            and '---' not in serialdata):
                         bdata['serial'] = serialdata
                     partnum = fundata.get(self.ADP_PARTNUM, None)
                     if partnum and partnum != 'N/A':
@@ -798,7 +822,7 @@ class IMMClient(object):
             try:
                 fpga = self.ipmicmd.xraw_command(netfn=0x3a, command=0x6b,
                                                  data=(0,))
-                fpga = '{0}.{1}.{2}'.format(*[ord(x) for x in fpga['data']])
+                fpga = '{0}.{1}.{2}'.format(*bytearray(fpga['data']))
                 yield ('FPGA', {'version': fpga})
             except pygexc.IpmiException as ie:
                 if ie.ipmicode != 193:
@@ -833,8 +857,16 @@ class XCCClient(IMMClient):
 
     def __init__(self, ipmicmd):
         super(XCCClient, self).__init__(ipmicmd)
+        self.ipmicmd.ipmi_session.register_keepalive(self.keepalive, None)
         self.adp_referer = None
 
+    def reseat(self):
+        rsp = self.wc.grab_json_response_with_status(
+            '/api/providers/virt_reseat', '{}')
+        if rsp[1] != 200 or rsp[0].get('return', 1) != 0:
+            raise pygexc.UnsupportedFunctionality(
+                'This platform does not support AC reseat.')
+
     def get_description(self):
         dsc = self.wc.grab_json_response('/DeviceDescription.json')
         dsc = dsc[0]
@@ -870,22 +902,38 @@ class XCCClient(IMMClient):
             'value': passrules.get('pass_min_length')}
         settings['password_lockout_period'] = {
             'value': passrules.get('lockout_period')}
+        presassert = self.wc.grab_json_response('/api/dataset/imm_rpp')
+        asserted = presassert.get('items', [{}])[0].get('rpp_Assert', None)
+        if asserted is not None:
+            settings['presence_assert'] = {
+                'value': 'Enable' if asserted else 'Disable'
+            }
         try:
             enclosureinfo = self.ipmicmd.xraw_command(0x3a, 0xf1, data=[0])
         except pygexc.IpmiException:
             return settings
         settings['smm'] = {
             'default': 'Disable',
-            'possible': ['Enable', 'Disable'],
-            'help': 'Enables or disables the network of the D2 '
-                    'enclosure manager.',
+            'possible': ['Enable', 'Disable', 'IPMI'],
+            'help': 'Enables or disables the network of the '
+                    'enclosure manager. IPMI Enables with IPMI '
+                    'for v2 systems.',
         }
         if bytearray(enclosureinfo['data'])[0] == 2:
             settings['smm']['value'] = 'Disable'
         elif bytearray(enclosureinfo['data'])[0] == 1:
             settings['smm']['value'] = 'Enable'
+        elif bytearray(enclosureinfo['data'])[0] == 4:
+            settings['smm']['value'] = 'IPMI'
         else:
             settings['smm']['value'] = None
+        smmip = self.get_property('/v2/ibmc/smm/smm_ip')
+        if smmip:
+            smmip = socket.inet_ntoa(bytes(smmip[-1::-1]))
+            settings['smm_ip'] = {
+                'help': 'Current IPv4 address as reported by SMM, read-only',
+                'value': smmip,
+            }
         return settings
 
     rulemap = {
@@ -905,15 +953,31 @@ class XCCClient(IMMClient):
                 changeset[key] = {'value': changeset[key]}
             currval = changeset[key].get('value', None)
             if 'smm'.startswith(key.lower()):
-                if currval == 'Enable':
+                if 'enabled'.startswith(currval.lower()):
                     self.ipmicmd.xraw_command(0x3a, 0xf1, data=[1])
-                elif currval == 'Disable':
+                elif 'disabled'.startswith(currval.lower()):
                     self.ipmicmd.xraw_command(0x3a, 0xf1, data=[2])
+                elif 'ipmi'.startswith(currval.lower()):
+                    self.ipmicmd.xraw_command(0x3a, 0xf1, data=[4])
             elif key.lower() in self.rulemap:
                 ruleset[self.rulemap[key.lower()]] = changeset[key]['value']
                 if key.lower() == 'password_expiration':
                     warntime = str(int(int(changeset[key]['value']) * 0.08))
                     ruleset['USER_GlobalPassExpWarningPeriod'] = warntime
+            elif 'presence_asserted'.startswith(key.lower()):
+                assertion = changeset[key]['value']
+                if 'enabled'.startswith(assertion.lower()):
+                    self.wc.grab_json_response('/api/dataset',
+                                               {'IMM_RPPAssert': '0'})
+                    self.wc.grab_json_response('/api/dataset',
+                                               {'IMM_RPPAssert': '1'})
+                elif 'disabled'.startswith(assertion.lower()):
+                    self.wc.grab_json_response('/api/dataset',
+                                               {'IMM_RPPAssert': '0'})
+                else:
+                    raise pygexc.InvalidParameterValue(
+                        '"{0}" is not a recognized value for {1}'.format(
+                            assertion, key))
             else:
                 raise pygexc.InvalidParameterValue(
                     '{0} not a known setting'.format(key))
@@ -925,9 +989,9 @@ class XCCClient(IMMClient):
             '/redfish/v1/Systems/1/Bios/Actions/Bios.ResetBios',
             {'Action': 'Bios.ResetBios'},
             headers={
-                'Authorization': 'Basic ' +
-                                 base64.b64encode(
-                                     self.username + ':' + self.password),
+                'Authorization': 'Basic %s' % base64.b64encode(
+                    (self.username + ':' + self.password).encode('utf8')
+                ).decode('utf8'),
                 'Content-Type': 'application/json'
             }
         )
@@ -952,12 +1016,16 @@ class XCCClient(IMMClient):
                             'password': self.password
                             })
         headers = {'Connection': 'keep-alive',
+                   'Referer': 'https://xcc/',
+                   'Host': 'xcc',
                    'Content-Type': 'application/json'}
         wc.request('POST', '/api/login', adata, headers)
         rsp = wc.getresponse()
         if rsp.status == 200:
             rspdata = json.loads(rsp.read())
             wc.set_header('Content-Type', 'application/json')
+            wc.set_header('Referer', 'https://xcc/')
+            wc.set_header('Host', 'xcc')
             wc.set_header('Authorization', 'Bearer ' + rspdata['access_token'])
             if '_csrf_token' in wc.cookies:
                 wc.set_header('X-XSRF-TOKEN', wc.cookies['_csrf_token'])
@@ -965,9 +1033,18 @@ class XCCClient(IMMClient):
 
     def _raid_number_map(self, controller):
         themap = {}
+        cid = controller.split(',')
         rsp = self.wc.grab_json_response(
             '/api/function/raid_conf?'
-            'params=raidlink_GetDisksToConf,{0}'.format(controller))
+            'params=raidlink_GetDisksToConf,{0}'.format(cid[0]))
+        if rsp.get('return') == 22:  # New style firmware
+            if cid[2] == 2:
+                arg = '{0},{1}'.format(cid[1], cid[2])
+            else:
+                arg = '{0},{1}'.format(cid[0], cid[2])
+            arg = 'params=raidlink_GetDisksToConf,{0}'.format(arg)
+            rsp = self.wc.grab_json_response(
+                '/api/function/raid_conf?{0}'.format(arg))
         for lvl in rsp['items'][0]['supported_raidlvl']:
             mapdata = (lvl['rdlvl'], lvl['maxSpan'])
             raidname = lvl['rdlvlstr'].replace(' ', '').lower()
@@ -1009,7 +1086,7 @@ class XCCClient(IMMClient):
                                                 {'Generate_FFDC_status': 1})
         url = '/ffdc/{0}'.format(result['FileName'])
         if autosuffix and not savefile.endswith('.tzz'):
-            savefile += '.tzz'
+            savefile += '-{0}'.format(result['FileName'])
         fd = webclient.FileDownloader(self.wc, url, savefile)
         fd.start()
         while fd.isAlive():
@@ -1018,6 +1095,8 @@ class XCCClient(IMMClient):
                 progress({'phase': 'download',
                           'progress': 100 * self.wc.get_download_progress()})
             self._refresh_token()
+        if fd.exc:
+            raise fd.exc
         if progress:
             progress({'phase': 'complete'})
         return savefile
@@ -1103,11 +1182,18 @@ class XCCClient(IMMClient):
             else:
                 hstr = ''
             drvstr = '|'.join([str(x.id[1]) for x in drives]) + '|'
+            ctl = controller.split(',')
             pth = '/api/function/raid_conf?params=raidlink_CheckConfisValid'
-            args = [pth, controller, rdlvl, spancount, drivesperspan, drvstr,
+            args = [pth, ctl[0], rdlvl, spancount, drivesperspan, drvstr,
                     hstr]
             url = ','.join([str(x) for x in args])
             rsp = self.wc.grab_json_response(url)
+            if rsp.get('return', -1) == 22:
+                args.append(ctl[1])
+                args = [pth, ctl[0], rdlvl, spancount, drivesperspan, drvstr,
+                        hstr, ctl[1]]
+                url = ','.join([str(x) for x in args])
+                rsp = self.wc.grab_json_response(url)
             if rsp['items'][0]['errcode'] == 16:
                 raise pygexc.InvalidParameterValue('Incorrect number of disks')
             elif rsp['items'][0]['errcode'] != 0:
@@ -1182,9 +1268,18 @@ class XCCClient(IMMClient):
         realcfg = self.get_storage_configuration(False)
         for pool in cfgspec.arrays:
             for volume in pool.volumes:
-                vid = '{0},{1}'.format(volume.id[1], volume.id[0])
+                cid = volume.id[0].split(',')
+                if cid[2] == 2:
+                    vid = '{0},{1},{2}'.format(volume.id[1], cid[1], cid[2])
+                else:
+                    vid = '{0},{1},{2}'.format(volume.id[1], cid[0], cid[2])
                 rsp = self.wc.grab_json_response(
                     '/api/function', {'raidlink_RemoveVolumeAsync': vid})
+                if rsp.get('return', -1) == 2:
+                    # older firmware
+                    vid = '{0},{1}'.format(volume.id[1], cid[0])
+                    rsp = self.wc.grab_json_response(
+                        '/api/function', {'raidlink_RemoveVolumeAsync': vid})
                 if rsp.get('return', -1) != 0:
                     raise Exception(
                         'Unexpected return to volume deletion: ' + repr(rsp))
@@ -1210,8 +1305,9 @@ class XCCClient(IMMClient):
 
     def _create_array(self, pool):
         params = self._parse_array_spec(pool)
+        cid = params['controller'].split(',')[0]
         url = '/api/function/raid_conf?params=raidlink_GetDefaultVolProp'
-        args = (url, params['controller'], 0, params['drives'])
+        args = (url, cid, 0, params['drives'])
         props = self.wc.grab_json_response(','.join([str(x) for x in args]))
         props = props['items'][0]
         volumes = pool.volumes
@@ -1236,18 +1332,26 @@ class XCCClient(IMMClient):
                     nameappend += 1
             else:
                 name = vol.name
-            if vol.stripsize:
+            if vol.stripsize is not None:
                 stripsize = int(math.log(vol.stripsize * 2, 2))
             else:
                 stripsize = props['stripsize']
+            if vol.read_policy is not None:
+                read_policy = vol.read_policy
+            else:
+                read_policy = props["cpra"]
+            if vol.write_policy is not None:
+                write_policy = vol.write_policy
+            else:
+                write_policy = props["cpwb"]
             strsize = 'remainder' if vol.size is None else str(vol.size)
             if strsize in ('all', '100%'):
                 volsize = params['capacity']
             elif strsize in ('remainder', 'rest'):
                 volsize = remainingcap
             elif strsize.endswith('%'):
-                volsize = int(params['capacity'] *
-                              float(strsize.replace('%', '')) / 100.0)
+                volsize = int(params['capacity']
+                              * float(strsize.replace('%', '')) / 100.0)
             else:
                 try:
                     volsize = int(strsize)
@@ -1259,15 +1363,26 @@ class XCCClient(IMMClient):
                 raise pygexc.InvalidParameterValue(
                     'Requested sizes exceed available capacity')
             vols.append('{0};{1};{2};{3};{4};{5};{6};{7};{8};|'.format(
-                name, volsize, stripsize, props['cpwb'], props['cpra'],
+                name, volsize, stripsize, write_policy, read_policy,
                 props['cpio'], props['ap'], props['dcp'], props['initstate']))
         url = '/api/function'
+        cid = params['controller'].split(',')
+        cnum = cid[0]
         arglist = '{0},{1},{2},{3},{4},{5},'.format(
-            params['controller'], params['raidlevel'], params['spans'],
+            cnum, params['raidlevel'], params['spans'],
             params['perspan'], params['drives'], params['hotspares'])
         arglist += ''.join(vols)
         parms = {'raidlink_AddNewVolWithNaAsync': arglist}
         rsp = self.wc.grab_json_response(url, parms)
+        if rsp['return'] == 14:  # newer firmware
+            if cid[2] == 2:
+                cnum = cid[1]
+            arglist = '{0},{1},{2},{3},{4},{5},'.format(
+                cnum, params['raidlevel'], params['spans'],
+                params['perspan'], params['drives'], params['hotspares'])
+            arglist += ''.join(vols) + ',{0}'.format(cid[2])
+            parms = {'raidlink_AddNewVolWithNaAsync': arglist}
+            rsp = self.wc.grab_json_response(url, parms)
         if rsp['return'] != 0:
             raise Exception(
                 'Unexpected response to add volume command: ' + repr(rsp))
@@ -1328,12 +1443,17 @@ class XCCClient(IMMClient):
 
     def get_storage_configuration(self, logout=True):
         rsp = self.wc.grab_json_response(
-            '/api/function/raid_alldevices?params=storage_GetAllDevices')
+            '/api/function/raid_alldevices?params=storage_GetAllDevices,0')
+        if not rsp:
+            rsp = self.wc.grab_json_response(
+                '/api/function/raid_alldevices?params=storage_GetAllDevices')
         standalonedisks = []
         pools = []
         for item in rsp.get('items', []):
             for cinfo in item['controllerInfo']:
-                cid = cinfo['id']
+                cid = '{0},{1},{2}'.format(
+                    cinfo['id'], cinfo.get('slotNo', -1), cinfo.get(
+                        'type', -1))
                 for pool in cinfo['pools']:
                     volumes = []
                     disks = []
@@ -1409,14 +1529,15 @@ class XCCClient(IMMClient):
                 raise pygexc.InvalidParameterValue(
                     'Given location was unreachable by the XCC')
             raise Exception('Unhandled return: ' + repr(rt))
-        if not self.webkeepalive:
+        if not self._keepalivesession:
             self._keepalivesession = self._wc
-            self.webkeepalive = self.ipmicmd.ipmi_session.register_keepalive(
-                self.keepalive, None)
         self._wc = None
 
     def keepalive(self):
-        self._refresh_token_wc(self._keepalivesession)
+        if self.fwo and util._monotonic_time() - self.fwovintage > 15:
+            self.fwo = None
+        if self._keepalivesession:
+            self._refresh_token_wc(self._keepalivesession)
 
     def fetch_psu_firmware(self):
         psudata = self.get_cached_data('lenovo_cached_psu')
@@ -1439,8 +1560,8 @@ class XCCClient(IMMClient):
         # then check for agentless, if agentless, get adapter info using
         # https, using the caller TLS verification scheme
         components = set(components)
-        if (not components or
-                set(('core', 'imm', 'bmc', 'xcc')) & components):
+        if (not components
+                or set(('core', 'imm', 'bmc', 'xcc')) & components):
             rsp = self.ipmicmd.xraw_command(netfn=0x3a, command=0x50)
             immverdata = self.parse_imm_buildinfo(rsp['data'])
             bmcmajor, bmcminor = [int(x) for x in bmcver.split('.')]
@@ -1448,20 +1569,20 @@ class XCCClient(IMMClient):
             bdata = {'version': bmcver,
                      'build': immverdata[0],
                      'date': immverdata[1]}
-            yield (self.bmcname, bdata)
+            yield self.bmcname, bdata
             bdata = self.fetch_grouped_properties({
                 'build': '/v2/ibmc/dm/fw/imm3/backup_pending_build_id',
                 'version': '/v2/ibmc/dm/fw/imm3/backup_pending_build_version',
                 'date': '/v2/ibmc/dm/fw/imm3/backup_pending_build_date'})
             if bdata:
-                yield ('{0} Backup'.format(self.bmcname), bdata)
+                yield '{0} Backup'.format(self.bmcname), bdata
             else:
                 bdata = self.fetch_grouped_properties({
                     'build': '/v2/ibmc/dm/fw/imm3/backup_build_id',
                     'version': '/v2/ibmc/dm/fw/imm3/backup_build_version',
                     'date': '/v2/ibmc/dm/fw/imm3/backup_build_date'})
                 if bdata:
-                    yield ('{0} Backup'.format(self.bmcname), bdata)
+                    yield '{0} Backup'.format(self.bmcname), bdata
                     bdata = self.fetch_grouped_properties({
                         'build': '/v2/ibmc/trusted_buildid',
                     })
@@ -1470,7 +1591,7 @@ class XCCClient(IMMClient):
                     'build': '/v2/ibmc/trusted_buildid',
                 })
             if bdata:
-                yield ('{0} Trusted Image'.format(self.bmcname), bdata)
+                yield '{0} Trusted Image'.format(self.bmcname), bdata
             bdata = self.fetch_grouped_properties({
                 'build': '/v2/ibmc/dm/fw/imm3/primary_pending_build_id',
                 'version': '/v2/ibmc/dm/fw/imm3/primary_pending_build_version',
@@ -1530,9 +1651,7 @@ class XCCClient(IMMClient):
                 yield firm
 
     def detach_remote_media(self):
-        if self.webkeepalive:
-            self.ipmicmd.ipmi_session.unregister_keepalive(self.webkeepalive)
-            self._keepalivesession = None
+        self._keepalivesession = None
         rt = self.wc.grab_json_response('/api/providers/rp_vm_remote_getdisk')
         if 'items' in rt:
             slots = []
@@ -1560,8 +1679,8 @@ class XCCClient(IMMClient):
                 url = mt['remotepath']
                 if url.startswith('//'):
                     url = 'smb:' + url
-                elif (not url.startswith('http://') and
-                      not url.startswith('https://')):
+                elif (not url.startswith('http://')
+                      and not url.startswith('https://')):
                     url = url.replace(':', '')
                     url = 'nfs://' + url
                 yield media.Media(mt['filename'], url)
@@ -1571,11 +1690,11 @@ class XCCClient(IMMClient):
                 yield media.Media(mt['filename'])
         self.weblogout()
 
-    def upload_media(self, filename, progress=None):
+    def upload_media(self, filename, progress=None, data=None):
         xid = random.randint(0, 1000000000)
         self._refresh_token()
         uploadthread = webclient.FileUploader(
-            self.wc, '/upload?X-Progress-ID={0}'.format(xid), filename, None)
+            self.wc, '/upload?X-Progress-ID={0}'.format(xid), filename, data)
         uploadthread.start()
         while uploadthread.isAlive():
             uploadthread.join(3)
@@ -1601,10 +1720,14 @@ class XCCClient(IMMClient):
         if rsp.get('return', -1) != 0:
             errmsg = repr(rsp) if rsp else self.wc.lastjsonerror
             raise Exception('Unrecognized return: ' + errmsg)
-        rsp = self.wc.grab_json_response('/api/providers/rp_rdoc_getfiles')
-        if 'items' not in rsp or len(rsp['items']) == 0:
-            raise Exception(
-                'Image upload was not accepted, it may be too large')
+        ready = False
+        while not ready:
+            ipmisession.Session.pause(3)
+            rsp = self.wc.grab_json_response('/api/providers/rp_rdoc_getfiles')
+            if 'items' not in rsp or len(rsp['items']) == 0:
+                raise Exception(
+                    'Image upload was not accepted, it may be too large')
+            ready = rsp['items'][0]['size'] != 0
         self._refresh_token()
         rsp = self.wc.grab_json_response('/api/providers/rp_rdoc_mountall',
                                          {})
@@ -1678,8 +1801,8 @@ class XCCClient(IMMClient):
                 progress({'phase': 'upload',
                           'progress': 100.0 * rsp['received'] / rsp['size']})
             elif rsp['state'] != 'done':
-                if (rsp.get('status', None) == 413 or
-                    uploadthread.rspstatus == 413):
+                if (rsp.get('status', None) == 413
+                        or uploadthread.rspstatus == 413):
                     raise Exception('File is larger than supported')
                 raise Exception('Unexpected result:' + repr(rsp))
             uploadstate = rsp['state']
@@ -1808,10 +1931,10 @@ class XCCClient(IMMClient):
 
         if rsp.get('return', -1) != 0:
             errmsg = repr(rsp) if rsp else self.wc.lastjsonerror
-            raise Exception('Unexpected result starting update: ' +
-                            errmsg)
+            raise Exception('Unexpected result starting update: %s' % errmsg)
         complete = False
         while not complete:
+            self._refresh_token()
             ipmisession.Session.pause(3)
             rsp = self.wc.grab_json_response(
                 '/api/dataset/imm_firmware_progress')
@@ -1825,8 +1948,8 @@ class XCCClient(IMMClient):
                 if rsp['items'][0]['action_status'] != 0:
                     raise Exception('Unexpected failure: ' + repr(rsp))
                 break
-            if (rsp['items'][0]['action_state'] == 'In Progress' and
-                    rsp['items'][0]['action_status'] == 2):
+            if (rsp['items'][0]['action_state'] == 'In Progress'
+                    and rsp['items'][0]['action_status'] == 2):
                 raise Exception('Unexpected failure: ' + repr(rsp))
             if rsp['items'][0]['action_state'] != 'In Progress':
                 raise Exception(
@@ -1864,7 +1987,7 @@ class XCCClient(IMMClient):
     def get_health(self, summary):
         try:
             wc = self.get_webclient(False)
-        except (socket.timeout, socket.error) as e:
+        except (socket.timeout, socket.error):
             wc = None
         if not wc:
             summary['health'] = pygconst.Health.Critical
@@ -1893,12 +2016,16 @@ class XCCClient(IMMClient):
             'E': pygconst.Health.Critical,
             'W': pygconst.Health.Warning,
         }
+        infoevents = False
         for item in rsp.get('items', ()):
             # while usually the ipmi interrogation shall explain things,
             # just in case there is a gap, make sure at least the
             # health field is accurately updated
             itemseverity = hmap.get(item.get('severity', 'E'),
                                     pygconst.Health.Critical)
+            if itemseverity == pygconst.Health.Ok:
+                infoevents = True
+                continue
             if (summary['health'] < itemseverity):
                 summary['health'] = itemseverity
             if item['cmnid'] == 'FQXSPPW0104J':
@@ -1926,7 +2053,8 @@ class XCCClient(IMMClient):
                     'health': itemseverity,
                     'type': item['source'],
                 }, ''))
-        if summary.get('health', pygconst.Health.Ok) == pygconst.Health.Ok:
+        if (summary.get('health', pygconst.Health.Ok) == pygconst.Health.Ok
+                and not infoevents):
             # Fault LED is lit without explanation, mark to encourage
             # examination
             summary['health'] = pygconst.Health.Warning
@@ -1975,7 +2103,7 @@ class XCCClient(IMMClient):
                     '/api/providers/imm_fod', {'FOD_LicenseKeyDelete': licid})
                 break
 
-    def apply_license(self, filename, progress=None):
+    def apply_license(self, filename, progress=None, data=None):
         license_errors = {
             310: "License is for a different model of system",
             311: "License is for a different system serial number",
@@ -1983,7 +2111,8 @@ class XCCClient(IMMClient):
             313: "License is expired",
             314: "License usage limit reached",
         }
-        uploadthread = webclient.FileUploader(self.wc, '/upload', filename)
+        uploadthread = webclient.FileUploader(self.wc, '/upload', filename,
+                                              data=data)
         uploadthread.start()
         uploadthread.join()
         rsp = json.loads(uploadthread.rsp)
