@@ -37,6 +37,24 @@ except NameError:
     pass
 
 
+psutypes = {
+    0x63: 'CFF v3 550W PT',
+    0x64: 'CFF v3 750W PT',
+    0x65: 'CFF v3 750W TT',
+    0x66: 'CFF v3 1100W PT',
+    0x67: 'CFF v3 1600W PT',
+    0x68: 'CFF v3 2000W PT',
+    0x6B: 'CFF v4 500W PT',
+    0x6C: 'CFF v4 750W PT',
+    0x6D: 'CFF v4 750W TT',
+    0x6E: 'CFF v4 1100W PT',
+    0x6F: 'CFF v4 1100W -48Vdc',
+    0x70: 'CFF v4 1100W 200-277Vac/240-380Vdc',
+    0x71: 'CFF v4 1800W PT',
+    0x72: 'CFF v4 2400W PT',
+}
+
+
 def fromstring(inputdata):
     if b'!entity' in inputdata.lower():
         raise Exception('!ENTITY not supported in this interface')
@@ -158,6 +176,33 @@ def fpc_read_powerbank(ipmicmd):
     return struct.unpack_from('<H', rsp['data'][3:])[0]
 
 
+def get_psu_count(ipmicmd, variant):
+    if variant == 0x26:
+        mymsg = ipmicmd.xraw_command(netfn=0x32, command=0xa8)
+        builddata = bytearray(mymsg['data'])
+        if builddata[13] == 3:
+            return 9
+        else:
+            return 6
+    else:
+        return variant & 0xf
+
+
+def fpc_get_dripstatus(ipmicmd, number, sz):
+    health = pygconst.Health.Ok
+    states = []
+    rsp = ipmicmd.xraw_command(0x34, 5)
+    rdata = bytearray(rsp['data'])
+    number = number - 1
+    if rdata[0] & (1 << number) == 0:
+        states.append('Absent')
+        health = pygconst.Health.Critical
+    if rdata[1] & (1 << number) != 0:
+        states.append('Leak detected')
+        health = pygconst.Health.Critical
+    return (health, states)
+
+
 fpc_sensors = {
     'AC Power': {
         'type': 'Power',
@@ -201,13 +246,22 @@ fpc_sensors = {
         'type': 'Power Supply',
         'returns': 'tuple',
         'units': None,
+        'elements': True,
         'provider': fpc_get_psustatus,
-        'elements': 1,
+        'elementsfun': get_psu_count,
+    },
+    'Drip Sensor': {
+        'type': 'Management Subsystem Health',
+        'elements': True,
+        'abselements': 2,
+        'returns': 'tuple',
+        'units': None,
+        'provider': fpc_get_dripstatus,
     }
 }
 
 
-def get_sensor_names(size):
+def get_sensor_names(ipmicmd, size):
     global fpc_sensors
     for name in fpc_sensors:
         if size != 6 and name in ('Fan Power', 'Total Power Capacity',
@@ -215,8 +269,19 @@ def get_sensor_names(size):
             continue
         if size == 6 and name == 'PSU Power Loss':
             continue
+        if size != 0x26 and name == 'Drip Sensor':
+            continue
         sensor = fpc_sensors[name]
-        if 'elements' in sensor:
+
+        if 'abselements' in sensor:
+            for elemidx in range(sensor['abselements']):
+                elemidx += 1
+                yield '{0} {1}'.format(name, elemidx)
+        elif 'elementsfun' in sensor:
+            for elemidx in range(sensor['elementsfun'](ipmicmd, size)):
+                elemidx += 1
+                yield '{0} {1}'.format(name, elemidx)
+        elif 'elements' in sensor:
             for elemidx in range(sensor['elements'] * (size & 0b11111)):
                 elemidx += 1
                 yield '{0} {1}'.format(name, elemidx)
@@ -224,7 +289,7 @@ def get_sensor_names(size):
             yield name
 
 
-def get_sensor_descriptions(size):
+def get_sensor_descriptions(ipmicmd, size):
     global fpc_sensors
     for name in fpc_sensors:
         if size != 6 and name in ('Fan Power', 'Total Power Capacity',
@@ -232,8 +297,20 @@ def get_sensor_descriptions(size):
             continue
         if size == 6 and name == 'PSU Power Loss':
             continue
+        if size != 0x26 and name == 'Drip Sensor':
+            continue
         sensor = fpc_sensors[name]
-        if 'elements' in sensor:
+        if 'abselements' in sensor:
+            for elemidx in range(sensor['abselements']):
+                elemidx += 1
+                yield {'name': '{0} {1}'.format(name, elemidx),
+                       'type': sensor['type']}
+        elif 'elementsfun' in sensor:
+            for elemidx in range(sensor['elementsfun'](ipmicmd, size)):
+                elemidx += 1
+                yield {'name': '{0} {1}'.format(name, elemidx),
+                       'type': sensor['type']}
+        elif 'elements' in sensor:
             for elemidx in range(sensor['elements'] * (size & 0b11111)):
                 elemidx += 1
                 yield {'name': '{0} {1}'.format(name, elemidx),
@@ -252,7 +329,7 @@ def get_fpc_firmware(bmcver, ipmicmd, fpcorsmm):
         else:
             name = 'SMM'
         buildid = '{0}{1}{2}{3}{4}{5}{6}'.format(
-            *[chr(x) for x in builddata[-7:]])
+            *[chr(x) for x in builddata[6:13]])
     elif len(builddata) == 8:
         builddata = builddata[1:]  # discard the 'completion code'
         name = 'FPC'
@@ -279,12 +356,20 @@ def get_sensor_reading(name, ipmicmd, sz):
     else:
         bnam, _, idx = name.rpartition(' ')
         idx = int(idx)
-        if bnam in fpc_sensors and idx <= fpc_sensors[bnam]['elements'] * sz:
-            sensor = fpc_sensors[bnam]
-            if 'returns' in sensor:
-                health, states = sensor['provider'](ipmicmd, idx, sz)
-            else:
-                value = sensor['provider'](ipmicmd, idx, sz)
+        if bnam in fpc_sensors:
+            max = -1
+            if 'abselements' in fpc_sensors[bnam]:
+                max = fpc_sensors[bnam]['abselements']
+            elif 'elementsfun' in fpc_sensors[bnam]:
+                max = 99
+            elif 'elements' in fpc_sensors[bnam]:
+                max = fpc_sensors[bnam]['elements'] * sz
+            if idx <= max:
+                sensor = fpc_sensors[bnam]
+                if 'returns' in sensor:
+                    health, states = sensor['provider'](ipmicmd, idx, sz)
+                else:
+                    value = sensor['provider'](ipmicmd, idx, sz)
     if sensor is not None:
         return sdr.SensorReading({'name': name, 'imprecision': None,
                                   'value': value, 'states': states,
@@ -296,7 +381,8 @@ def get_sensor_reading(name, ipmicmd, sz):
 
 class SMMClient(object):
 
-    def __init__(self, ipmicmd):
+    def __init__(self, ipmicmd, variant):
+        self.smm_variant = variant
         self.ipmicmd = weakref.proxy(ipmicmd)
         self.smm = ipmicmd.bmc
         self.username = ipmicmd.ipmi_session.userid
@@ -341,7 +427,7 @@ class SMMClient(object):
                 settings[rule] = {'value': int(ruleinfo.text)}
         dwc = self.ipmicmd.xraw_command(0x32, 0x94)
         dwc = bytearray(dwc['data'])
-        if len(dwc) != 3 or dwc[0] == 1:
+        if len(dwc) not in (3, 4) or dwc[0] == 1:
             rsp = self.ipmicmd.xraw_command(0x34, 3)
             fanmode = self.fanmodes[bytearray(rsp['data'])[0]]
             settings['fanspeed'] = {
@@ -473,7 +559,7 @@ class SMMClient(object):
             self.wc.request('POST', '/data', rules)
             self.wc.getresponse().read()
         if powercfg != [None, None]:
-            if variant == 2:
+            if variant != 6:
                 if None in powercfg:
                     currcfg = self.ipmicmd.xraw_command(0x32, 0xa2)
                     currcfg = bytearray(currcfg['data'])
@@ -498,8 +584,10 @@ class SMMClient(object):
         if priv.lower() == 'administrator':
             rsp = self.ipmicmd.xraw_command(netfn=6, command=0x46, data=(uid,))
             username = bytes(rsp['data']).rstrip(b'\x00')
+            if not isinstance(username, str):
+                username = username.decode('utf8')
             self.wc.request(
-                'POST', '/data', b'set=user({0},1,{1},511,,4,15,0)'.format(
+                'POST', '/data', 'set=user({0},1,{1},511,,4,15,0)'.format(
                     uid, username))
             rsp = self.wc.getresponse()
             rsp.read()
@@ -508,7 +596,8 @@ class SMMClient(object):
         self.ipmicmd.xraw_command(netfn=0x32, command=0xa4,
                                   data=[int(bay), 2])
 
-    def get_diagnostic_data(self, savefile, progress=None, autosuffix=False):
+    def get_diagnostic_data(self, savefile, progress=None, autosuffix=False,
+                            variant=None):
         rsp = self.ipmicmd.xraw_command(netfn=0x32, command=0xb1, data=[0])
         if bytearray(rsp['data'])[0] != 0:
             raise Exception("Service data generation already in progress")
@@ -529,7 +618,10 @@ class SMMClient(object):
                 progress({'phase': 'initializing', 'progress': initpct})
         if self.wc is None:
             raise Exception("Failed to connect to web api")
-        url = '/preview/smm-ffdc.tgz?ST1={0}'.format(self.st1)
+        if variant and variant >> 5:
+            url = '/preview/smm2-ffdc.tgz?ST1={0}'.format(self.st1)
+        else:
+            url = '/preview/smm-ffdc.tgz?ST1={0}'.format(self.st1)
         if autosuffix and not savefile.endswith('.tgz'):
             savefile += '-smm-ffdc.tgz'
         fd = webclient.FileDownloader(self.wc, url, savefile)
@@ -544,13 +636,17 @@ class SMMClient(object):
         return savefile
 
     def process_fru(self, fru):
+        smmv1 = self.smm_variant & 0xf0 == 0
         # TODO(jjohnson2): can also get EIOM, SMM, and riser data if warranted
-        fru['Serial Number'] = bytes(self.ipmicmd.xraw_command(
-            netfn=0x32, command=0xb0, data=(5, 1))['data'][:]).strip(
-                b' \x00\xff').replace(b'\xff', b'')
-        fru['Model'] = bytes(self.ipmicmd.xraw_command(
-            netfn=0x32, command=0xb0, data=(5, 0))['data'][:]).strip(
-                b' \x00\xff').replace(b'\xff', b'')
+        snum = bytes(self.ipmicmd.xraw_command(
+            netfn=0x32, command=0xb0, data=(5, 1))['data'][:])
+        mnum = bytes(self.ipmicmd.xraw_command(
+            netfn=0x32, command=0xb0, data=(5, 0))['data'][:])
+        if not smmv1:
+            snum = snum[2:]
+            mnum = mnum[2:]
+        fru['Serial Number'] = snum.strip(b' \x00\xff').replace(b'\xff', b'')
+        fru['Model'] = mnum.strip(b' \x00\xff').replace(b'\xff', b'')
         return fru
 
     def get_webclient(self):
@@ -767,6 +863,42 @@ class SMMClient(object):
             complete = percent >= 100.0
         return 'complete'
 
+    def get_inventory_descriptions(self, ipmicmd, variant):
+        psucount = get_psu_count(ipmicmd, variant)
+        for idx in range(psucount):
+            yield 'PSU {}'.format(idx + 1)
+
+    def get_inventory_of_component(self, ipmicmd, component):
+        psuidx = int(component.replace('PSU ', ''))
+        return self.get_psu_info(ipmicmd, psuidx)
+
+    def get_psu_info(self, ipmicmd, psunum):
+        psuinfo = ipmicmd.xraw_command(0x34, 0x6, data=(psunum,))
+        psuinfo = bytearray(psuinfo['data'])
+        psutype = struct.unpack('<H', psuinfo[23:25])[0]
+        psui = {}
+        if psuinfo[0] != 1:
+            return {'Model': 'Unavailable'}
+        psui['Revision'] = psuinfo[34]
+        psui['Description'] = psutypes.get(
+            psutype, 'Unknown ({})'.format(psutype))
+        psui['Part Number'] = str(psuinfo[35:47].strip().decode('utf8'))
+        psui['FRU Number'] = str(psuinfo[47:59].strip().decode('utf8'))
+        psui['Serial Number'] = str(psuinfo[59:71].strip().decode('utf8'))
+        psui['Header Code'] = str(psuinfo[71:75].strip().decode('utf8'))
+        psui['Vendor'] = str(psuinfo[25:29].strip().decode('utf8'))
+        psui['Manufacturing Date'] = '20{}-W{}'.format(
+            psuinfo[77:79].decode('utf8'), psuinfo[75:77].decode('utf8'))
+        psui['Primary Firmware Version'] = '{:x}.{:x}'.format(
+            psuinfo[80], psuinfo[79])
+        psui['Secondary Firmware Version'] = '{:x}.{:x}'.format(
+            psuinfo[82], psuinfo[81])
+        psui['Model'] = str(psuinfo[5:23].strip().decode('utf8'))
+        psui['Manufacturer Location'] = str(
+            psuinfo[83:85].strip().decode('utf8'))
+        psui['Barcode'] = str(psuinfo[85:108].strip().decode('utf8'))
+        return psui
+
     def logout(self):
         self.wc.request('POST', '/data/logout', None)
         rsp = self.wc.getresponse()
