@@ -1110,6 +1110,10 @@ class Command(object):
                     if fnmatch(attr.lower(), change.lower()):
                         found = True
                         changeset[attr] = changeset[change]
+                    if fnmatch(attr.lower(),
+                               change.replace('.', '_').lower()):
+                        found = True
+                        changeset[attr] = changeset[change]
                 if found:
                     del changeset[change]
         for change in changeset:
@@ -1168,11 +1172,13 @@ class Command(object):
                 netmask = _cidr_to_mask(int(cidr))
             patch['IPv4StaticAddresses'] = [ipinfo]
             ipinfo['Address'] = ipv4_address
+            ipv4_configuration = 'static'
             if netmask:
                 ipinfo['SubnetMask'] = netmask
         if ipv4_gateway:
             patch['IPv4StaticAddresses'] = [ipinfo]
             ipinfo['Gateway'] = ipv4_gateway
+            ipv4_configuration = 'static'
         if ipv4_configuration.lower() == 'dhcp':
             dodhcp = True
             patch['DHCPv4'] = {'DHCPEnabled': True}
@@ -1450,7 +1456,7 @@ class Command(object):
             return
         for vmurl in vmurls:
             vminfo = self._do_web_request(vmurl, cache=False)
-            if vminfo['ConnectedVia'] != 'NotConnected':
+            if vminfo.get('ConnectedVia', None) != 'NotConnected':
                 continue
             self._do_web_request(vmurl, {'Image': url, 'Inserted': True},
                                  'PATCH')
