@@ -95,6 +95,16 @@ class OEMHandler(generic.OEMHandler):
         self.updating = False
         self.datacache = {}
 
+    def reseat_bay(self, bay):
+        if bay != -1:
+            raise pygexc.UnsupportedFunctionality(
+                'This is not an enclosure manager')
+        rsp = self.wc.grab_json_response_with_status(
+            '/api/providers/virt_reseat', '{}')
+        if rsp[1] != 200 or rsp[0].get('return', 1) != 0:
+            raise pygexc.UnsupportedFunctionality(
+                'This platform does not support AC reseat.')
+
     def get_cached_data(self, attribute, age=30):
         try:
             kv = self.datacache[attribute]
@@ -656,11 +666,11 @@ class OEMHandler(generic.OEMHandler):
             raise pygexc.InvalidParameterValue(
                 'XCC does not have required license for operation')
 
-    def upload_media(self, filename, progress=None):
+    def upload_media(self, filename, progress=None, data=None):
         xid = random.randint(0, 1000000000)
         self._refresh_token()
         uploadthread = webclient.FileUploader(
-            self.wc, '/upload?X-Progress-ID={0}'.format(xid), filename, None)
+            self.wc, '/upload?X-Progress-ID={0}'.format(xid), filename, data)
         uploadthread.start()
         while uploadthread.isAlive():
             uploadthread.join(3)
@@ -918,7 +928,7 @@ class OEMHandler(generic.OEMHandler):
                                                 {'Generate_FFDC_status': 1})
         url = '/ffdc/{0}'.format(result['FileName'])
         if autosuffix and not savefile.endswith('.tzz'):
-            savefile += '.tzz'
+            savefile += '-{0}'.format(result['FileName'])
         fd = webclient.FileDownloader(self.wc, url, savefile)
         fd.start()
         while fd.isAlive():
